# SPDX-License-Identifier: Apache-2.0
#
# This file is part of the M2-ISA-R project: https://github.com/tum-ei-eda/M2-ISA-R
#
# Copyright (C) 2022
# Chair of Electrical Design Automation
# Technical University of Munich

"""Filter M2-ISA-R/Seal5 metamodel."""

import re
import sys
import argparse
import logging
import pathlib
import pickle
from typing import Union

from m2isar.metamodel import arch

logger = logging.getLogger("filter_model")


class DropUnusedContext:
    def __init__(self, names: "list[str]"):
        self.names = names
        self.to_keep = set()

    @property
    def to_drop(self):
        return set(name for name in self.names if name not in self.to_keep)

    def track(self, name: str):
        if name in self.names:
            # logger.debug("Tracked use of %s", name)
            self.to_keep.add(name)


def get_parser():
    # read command line args
    parser = argparse.ArgumentParser()
    parser.add_argument("top_level", help="A .m2isarmodel or .seal5model file.")
    parser.add_argument("--keep-sets", type=str, default=None)
    parser.add_argument("--drop-sets", type=str, default=None)
    parser.add_argument("--keep-instructions", type=str, default=None)
    parser.add_argument("--drop-instructions", type=str, default=None)
    parser.add_argument("--keep-opcodes", type=str, default=None)
    parser.add_argument("--drop-opcodes", type=str, default=None)
    parser.add_argument("--keep-encoding-sizes", type=str, default=None)
    parser.add_argument("--drop-encoding-sizes", type=str, default=None)
    # TODO: filter builtins/aliases
    parser.add_argument("--log", default="info", choices=["critical", "error", "warning", "info", "debug"])
    parser.add_argument("--output", "-o", type=str, default=None)
    return parser


def run(args):
    # initialize logging
    logging.basicConfig(level=getattr(logging, args.log.upper()))

    # resolve model paths
    top_level = pathlib.Path(args.top_level)
    # abs_top_level = top_level.resolve()

    is_seal5_model = False
    if args.output is None:  # inplace
        assert top_level.suffix in [".m2isarmodel", ".seal5model"], "Can not infer model type from file extension."
        if top_level.suffix == ".seal5model":
            is_seal5_model = True

        model_path = top_level
    else:
        model_path = pathlib.Path(args.output)
        assert top_level.suffix in [".m2isarmodel", ".seal5model"], "Can not infer model type from file extension."
        if top_level.suffix == ".seal5model":
            is_seal5_model = True

    logger.info("loading models")
    if not is_seal5_model:
        raise NotImplementedError

    # load models
    with open(top_level, "rb") as f:
        # models: "dict[str, arch.CoreDef]" = pickle.load(f)
        if is_seal5_model:
            model: "dict[str, Union[arch.InstructionSet, ...]]" = pickle.load(f)
            model["cores"] = {}
        else:  # TODO: core vs. set!
            temp: "dict[str, Union[arch.InstructionSet, arch.CoreDef]]" = pickle.load(f)
            assert len(temp) > 0, "Empty model!"
            if isinstance(list(temp.values())[0], arch.CoreDef):
                model = {"cores": temp, "sets": {}}
            elif isinstance(list(temp.values())[0], arch.InstructionSet):
                model = {"sets": temp, "cores": {}}
            else:
                assert False

    # preprocess model
    # print("model", model["sets"]["XCoreVMac"].keys())
    if args.keep_sets:
        keep_sets = args.keep_sets.split(",")
    else:
        keep_sets = []
    if args.drop_sets:
        drop_sets = args.drop_sets.split(",")
    else:
        drop_sets = []
    if args.keep_instructions:
        keep_instructions = args.keep_instructions.split(",")
    else:
        keep_instructions = []
    if args.drop_instructions:
        drop_instructions = args.drop_instructions.split(",")
    else:
        drop_instructions = []

    def opcodes_helper(x):
        OPCODE_LOOKUP = {
            "LOAD": 0b00000,
            "LOAD-FP": 0b00001,
            "custom-0": 0b00010,
            "MISC-MEM": 0b00011,
            "OP-IMM": 0b00100,
            "AUIPC": 0b00101,
            "OP-IMM-32": 0b00110,
            # "48bit": 0b00111,
            "STORE": 0b01000,
            "STORE-FP": 0b01001,
            "custom-1": 0b01010,
            "AMO": 0b01011,
            "OP": 0b01100,
            "LUI": 0b01101,
            "OP-32": 0b01110,
            # "64bit": 0b01111,
            "MADD": 0b10000,
            "MSUB": 0b10001,
            "NMADD": 0b10010,
            "NMSUB": 0b10011,
            "OP-FP": 0b10100,
            "OP-V": 0b10101,
            "custom-2": 0b10110,  # rv128i
            # "48bit2": 0b10111,
            "BRANCH": 0b11000,
            "JALR": 0b11001,
            # "reserved": 0b11010,
            "JAL": 0b11011,
            "SYSTEM": 0b11100,
            "OP-P": 0b11101,
            "custom-3": 0b11110,
            # "80bit+": 0b11111,
        }
        try:
            x = int(x, 0)
        except ValueError:
            x = OPCODE_LOOKUP.get(x, None)
            if x is None:
                raise RuntimeError(f"Opcode lookup by string failed for {x}. Options: {list(OPCODE_LOOKUP.keys())}")
        assert x >= 0
        assert x < 2**5
        return x

    if args.keep_opcodes:
        keep_opcodes = args.keep_opcodes.split(",")
        keep_opcodes = list(map(opcodes_helper, keep_opcodes))
    else:
        keep_opcodes = []
    if args.drop_opcodes:
        drop_opcodes = args.drop_opcodes.split(",")
        drop_opcodes = list(map(opcodes_helper, drop_opcodes))
    else:
        drop_opcodes = []
    if args.keep_encoding_sizes:
        keep_encoding_sizes = args.keep_encoding_sizes.split(",")
        keep_encoding_sizes = list(map(int, keep_encoding_sizes))
    else:
        keep_encoding_sizes = []
    if args.drop_encoding_sizes:
        drop_encoding_sizes = args.drop_encoding_sizes.split(",")
        drop_encoding_sizes = list(map(int, drop_encoding_sizes))
    else:
        drop_encoding_sizes = []
    # print("keep", keep_instructions)
    # print("drop", drop_instructions)
    # input("456")
    # def check_filter(name, keep, drop):
    #     if drop and keep:
    #         return name not in drop and name in keep
    #     elif keep:
    #         return name in keep
    #     elif drop:
    #         return name not in drop
    #     return True

    def check_filter_regex(name, keep, drop):
        if drop and keep:
            return not any(re.compile(expr).match(name) for expr in drop) and any(
                re.compile(expr).match(name) for expr in keep
            )
        elif keep:
            return any(re.compile(expr).match(name) for expr in keep)
        elif drop:
            return not any(re.compile(expr).match(name) for expr in drop)
        return True

    def check_encoding_filter(name, enc, keep, drop, keep2, drop2):
        opcode = None
        size = 0
        for e in reversed(enc):
            # print("e", e, dir(e))
            if isinstance(e, arch.BitVal):
                length = e.length
                if size == 0:
                    if length == 7:
                        val = e.value
                        opcode = val >> 2
            elif isinstance(e, arch.BitField):
                length = e.range.length
            else:
                assert False
            size += length
        assert size in [16, 32, 64, 128], f"Invalid size: {size} (Instruction: {name})"
        ret = True
        if drop2 and keep2:
            ret = size not in drop2 and size in keep2
        elif keep2:
            ret = size in keep2
        elif drop2:
            ret = size not in drop2
        if not ret:
            return False
        if opcode is None:  # not found (not a riscv insn?)
            return True
        if drop and keep:
            return opcode not in drop and opcode in keep
        elif keep:
            return opcode in keep
        elif drop:
            return opcode not in drop
        return True

    model["sets"] = {
        set_name: set_def
        for set_name, set_def in model["sets"].items()
        if check_filter_regex(set_name, keep_sets, drop_sets)
    }
    for set_name, set_def in model["sets"].items():
        set_def.instructions = {
            key: instr_def
            for key, instr_def in set_def.instructions.items()
            if check_filter_regex(instr_def.name, keep_instructions, drop_instructions)
            and check_encoding_filter(
                instr_def.name, instr_def.encoding, keep_opcodes, drop_opcodes, keep_encoding_sizes, drop_encoding_sizes
            )
        }
        # for instr_name, instr_def in set_def.instructions.items():

    # Remove sets without instructions
    model["sets"] = {set_name: set_def for set_name, set_def in model["sets"].items() if len(set_def.instructions) > 0}

    logger.info("dumping model")
    with open(model_path, "wb") as f:
        if is_seal5_model:
            pickle.dump(model, f)
        else:
            if len(model["sets"]) > 0:
                assert len(model["cores"]) == 0
                pickle.dump(model["sets"], f)
            elif len(model["cores"]) > 0:
                assert len(model["sets"]) == 0
                pickle.dump(model["cores"], f)
            else:
                assert False


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    run(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2023 TUM Department of Electrical and Computer Engineering.
#
# This file is part of Seal5.
# See https://github.com/tum-ei-eda/seal5.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Settings module for seal5."""
import logging
from pathlib import Path
from dataclasses import dataclass, field, asdict, fields
from typing import List, Union, Optional, Dict

import yaml
from dacite import from_dict

from seal5.types import PatchStage


DEFAULT_SETTINGS = {
    # "directory": ?,
    "name": "default",
    "logging": {
        "console": {
            "level": "INFO",
        },
        "file": {
            "level": "DEBUG",
            "rotate": False,
            "limit": 1000,
        },
    },
    "git": {
        "author": "Seal5",
        "mail": "example@example.com",
        "prefix": "[Seal5]",
    },
    "patches": [
        {
            "name": "gitignore",
            "file": "gitignore.patch",
            "target": "llvm",
            "stage": 0,
            "comment": "Add .seal5 directory to .gitignore",
        },
    ],
    "filter": {
        "sets": {
            "keep": [],
            "drop": [],
        },
        "instructions": {
            "keep": [],
            "drop": [],
        },
        "aliases": {
            "keep": [],
            "drop": [],
        },
        "intrinsics": {
            "keep": [],
            "drop": [],
        },
        "opcodes": {
            "keep": [],
            "drop": [],
        },
        "encoding_sizes": {
            "keep": [],
            "drop": [],
        },
    },
    # "transform": {
    #     "passes": "*",
    # },
    "passes": {
        "defaults": {
            "skip": [],
            "only": [],
            "overrides": {},
        },
        "per_model": {},
    },
    "test": {
        "paths": [],
        # "paths": ["MC/RISCV", "CodeGen/RISCV"],
    },
    "llvm": {
        "state": {"version": "auto", "base_commit": "unknown"},
        "ninja": True,
        "default_config": "release",
        "clone_depth": None,
        "configs": {
            "release": {
                "options": {
                    "CMAKE_BUILD_TYPE": "Release",
                    "LLVM_BUILD_TOOLS": True,
                    "LLVM_ENABLE_ASSERTIONS": False,
                    "LLVM_OPTIMIZED_TABLEGEN": True,
                    "LLVM_ENABLE_PROJECTS": ["clang", "lld"],
                    "LLVM_TARGETS_TO_BUILD": ["RISCV"],
                },
            },
            "release_assertions": {
                "options": {
                    "CMAKE_BUILD_TYPE": "Release",
                    "LLVM_BUILD_TOOLS": True,
                    "LLVM_ENABLE_ASSERTIONS": True,
                    "LLVM_OPTIMIZED_TABLEGEN": True,
                    "LLVM_ENABLE_PROJECTS": ["clang", "lld"],
                    "LLVM_TARGETS_TO_BUILD": ["RISCV"],
                },
            },
            "debug": {
                "options": {
                    "CMAKE_BUILD_TYPE": "Debug",
                    "LLVM_BUILD_TOOLS": True,
                    "LLVM_ENABLE_ASSERTIONS": True,
                    "LLVM_OPTIMIZED_TABLEGEN": True,
                    "LLVM_ENABLE_PROJECTS": ["clang", "lld"],
                    "LLVM_TARGETS_TO_BUILD": ["RISCV"],
                },
            },
        },
    },
    "inputs": [],
    "extensions": {
        # RV32Zpsfoperand:
        #   feature: RV32Zpsfoperand
        #   arch: rv32zpsfoperand
        #   version: "1.0"
        #   experimental: true
        #   vendor: false
        #   instructions/intrinsics/aliases/constraints: TODO
        #   # patches: []
    },
    "groups": {
        "all": "*",
    },
    "tools": {
        "pattern_gen": {
            "integrated": True,
            "clone_url": None,
            "ref": None,
            # "clone_depth": None,
            "clone_depth": 1,
            "sparse_checkout": True,
        },
    },
}


def check_supported_types(data):
    ALLOWED_TYPES = (int, float, str, bool)
    if isinstance(data, dict):
        for value in data.values():
            check_supported_types(value)
    elif isinstance(data, list):
        for x in data:
            check_supported_types(x)
    else:
        if data is not None:
            assert isinstance(data, ALLOWED_TYPES), f"Unsupported type: {type(data)}"


class YAMLSettings:
    @classmethod
    def from_dict(cls, data: dict):
        # print("from_dict", data)
        return from_dict(data_class=cls, data=data)

    @classmethod
    def from_yaml(cls, text: str):
        data = yaml.safe_load(text)
        return cls.from_dict(data)

    @classmethod
    def from_yaml_file(cls, path: Path):
        with open(path, "r") as file:
            data = yaml.safe_load(file)
        return cls.from_dict(data=data)

    def to_yaml(self):
        data = asdict(self)
        check_supported_types(data)
        text = yaml.dump(data)
        return text

    def to_yaml_file(self, path: Path):
        text = self.to_yaml()
        with open(path, "w") as file:
            file.write(text)

    def merge(self, other: "YAMLSettings", overwrite: bool = False):
        # print("merge", type(self), type(other))
        for f1 in fields(other):
            k1 = f1.name
            # if k1 == "prefix":
            #     input("1 PREFIX")
            v1 = getattr(other, k1)
            if v1 is None:
                # print("cont1")
                continue
            t1 = type(v1)
            # print("f1", f1.name, f1, type(f1))
            # print("other", k1, v1)
            found = False
            for f2 in fields(self):
                k2 = f2.name
                # if k2 == "prefix":
                #     print("f1", f1)
                #     print("f2", f2)
                #     input("2 PREFIX")
                v2 = getattr(self, k2)
                if k2 == k1:
                    found = True
                    if v2 is None:
                        setattr(self, k2, v1)
                    else:
                        t2 = type(v2)
                        assert t1 is t2, "Type conflict"
                        if isinstance(v1, YAMLSettings):
                            v2.merge(v1, overwrite=overwrite)
                        elif isinstance(v1, dict):
                            if overwrite:
                                v2.clear()
                                v2.update(v1)
                            else:
                                for dict_key, dict_val in v1.items():
                                    if dict_key in v2:
                                        if isinstance(dict_val, YAMLSettings):
                                            assert isinstance(v2[dict_key], YAMLSettings)
                                            v2[dict_key].merge(dict_val, overwrite=overwrite)
                                    else:
                                        v2[dict_key] = dict_val
                        elif isinstance(v1, list):
                            if overwrite:
                                v2.clear()
                            # duplicates are dropped here
                            new = [x for x in v1 if x not in v2]
                            v2.extend(new)
                        else:
                            assert isinstance(
                                v2, (int, float, str, bool, Path)
                            ), f"Unsupported field type for merge {t1}"
                            setattr(self, k1, v1)
                    break
                # print("f2", f2.name, f2, type(f2))
                # print("self", k2, v2)
            assert found

        # input("123")
        # if overwrite:
        #     self.data.update(other.data)
        # else:
        #     # self.data = utils.merge_dicts(self.data, other.data)

    # @staticmethod
    # def from_yaml(text: str):
    #     data = yaml.safe_load(text)
    #     return Seal5Settings(data=data)

    # @staticmethod
    # def from_yaml_file(path: Path):
    #     with open(path, "r") as file:
    #         data = yaml.safe_load(file)
    #     return Seal5Settings(data=data)

    # # def __init__(self, data: dict = {}):
    # #     self.data: dict = data
    # #     assert self.validate()

    # def to_yaml(self):
    #     data = self.data
    #     text = yaml.dump(data)
    #     return text

    # def to_yaml_file(self, path: Path):
    #     text = self.to_yaml()
    #     with open(path, "w") as file:
    #         file.write(text)

    # def validate(self):
    #     # TODO
    #     return True

    # def merge(self, other: "YAMLSettings", overwrite: bool = False):
    #     # TODO:
    #     if overwrite:
    #         self.data.update(other.data)
    #     else:
    #         self.data = utils.merge_dicts(self.data, other.data)


@dataclass
class TestSettings(YAMLSettings):
    paths: Optional[List[Union[Path, str]]] = None


@dataclass
class GitSettings(YAMLSettings):
    author: str = "Seal5"
    mail: str = "example@example.com"
    prefix: Optional[str] = "[Seal5]"


@dataclass
class PatchSettings(YAMLSettings):
    name: str
    target: Optional[str] = None
    stage: Optional[Union[PatchStage, int]] = None  # TODO: default to 0? Allow int with union?
    comment: Optional["str"] = None
    file: Optional[Union[Path, str]] = None
    index: Optional[Union[Path, str]] = None
    # _file: Optional[Union[Path, str]] = field(init=False, repr=False)
    enable: bool = True
    generated: bool = False
    applied: bool = False

    # @property
    # def file(self) -> Path:
    #     return self._file

    # @file.setter
    # def file(self, value: Optional[Union[Path, str]]):
    #     self._file = value


# @dataclass
# class TransformSettings(YAMLSettings):
#     pass


@dataclass
class PassesSetting(YAMLSettings):
    skip: Optional[List[str]] = None
    only: Optional[List[str]] = None
    overrides: Optional[dict] = None


@dataclass
class PassesSettings(YAMLSettings):
    defaults: Optional[PassesSetting] = None
    per_model: Optional[Dict[str, PassesSetting]] = None


@dataclass
class ExtensionsSettings(YAMLSettings):
    feature: Optional[str] = None
    predicate: Optional[str] = None
    arch: Optional[str] = None
    version: Optional[str] = None
    experimental: Optional[bool] = None
    vendor: Optional[bool] = None
    std: Optional[bool] = None
    model: Optional[str] = None
    description: Optional[str] = None
    requires: Optional[List[str]] = None
    instructions: Optional[List[str]] = None
    # patches

    def get_version(self):
        if self.version:
            return self.version
        else:
            return "1.0"

    def get_description(self, name: Optional[str] = None):
        if self.description is None:
            if name:
                description = name
            else:
                description = "RISC-V"
            description += " Extension"
            return description
        else:
            return self.description

    def get_arch(self, name: Optional[str] = None):
        if self.arch is None:
            feature = self.get_feature(name=name)
            assert feature is not None
            arch = feature.lower()
            if self.experimental:
                arch = "experimental-" + arch
            return arch
        return self.arch

    def get_feature(self, name: Optional[str] = None):
        if self.feature is None:
            assert name is not None
            feature = name.replace("_", "")
            return feature
        else:
            return self.feature

    def get_predicate(self, name: Optional[str] = None, with_has: bool = False):
        if self.predicate is None:
            feature = self.get_feature(name=name)
            assert feature is not None
            if self.vendor:
                assert not self.std
                prefix = "Vendor"
            elif self.std:
                prefix = "StdExt"
            else:
                prefix = "Ext"
            if with_has:
                prefix = "Has" + prefix
            return prefix + feature
        else:
            if with_has:
                assert "has" in self.predicate.lower()
            return self.predicate


@dataclass
class GroupsSettings(YAMLSettings):
    pass


@dataclass
class ConsoleLoggingSettings(YAMLSettings):
    level: Union[int, str] = logging.INFO


@dataclass
class FileLoggingSettings(YAMLSettings):
    level: Union[int, str] = logging.INFO
    limit: Optional[int] = None  # TODO: implement
    rotate: bool = False  # TODO: implement


@dataclass
class LoggingSettings(YAMLSettings):
    console: ConsoleLoggingSettings
    file: FileLoggingSettings


@dataclass
class FilterSetting(YAMLSettings):
    keep: List[Union[str, int]] = field(default_factory=list)
    drop: List[Union[str, int]] = field(default_factory=list)


@dataclass
class FilterSettings(YAMLSettings):
    sets: Optional[FilterSetting] = None
    instructions: Optional[FilterSetting] = None
    aliases: Optional[FilterSetting] = None
    intrinsics: Optional[FilterSetting] = None
    opcodes: Optional[FilterSetting] = None
    encoding_sizes: Optional[FilterSetting] = None
    # TODO: functions


@dataclass
class LLVMVersion(YAMLSettings):
    major: Optional[int] = None
    minor: Optional[int] = None
    patch: Optional[int] = None
    rc: Optional[int] = None


@dataclass
class LLVMState(YAMLSettings):
    base_commit: Optional[str] = None
    version: Optional[Union[str, LLVMVersion]] = None


@dataclass
class LLVMConfig(YAMLSettings):
    options: dict = field(default_factory=dict)


@dataclass
class LLVMSettings(YAMLSettings):
    ninja: Optional[bool] = None
    clone_depth: Optional[int] = None
    default_config: Optional[str] = None
    configs: Optional[Dict[str, LLVMConfig]] = None
    state: Optional[LLVMState] = None


@dataclass
class RISCVLegalizerSetting(YAMLSettings):
    name: Optional[Union[str, List[str]]] = None
    types: Optional[Union[str, List[str]]] = None
    onlyif: Optional[Union[str, List[str]]] = None


@dataclass
class RISCVLegalizerSettings(YAMLSettings):
    ops: Optional[List[RISCVLegalizerSetting]] = None


@dataclass
class RISCVSettings(YAMLSettings):
    xlen: Optional[int] = None
    features: Optional[List[str]] = None
    # Used for  baseline extensions, tune and legalizer + pattern gen
    # default: zicsr,m,(32/64bit),fast-unaligned-access
    # others: zmmul,a,f,d,zfh,zfinx,zdinx,c,zba,zbb,zbc,zbs,zca,zcb,zcd
    #   zcmp,zce,e,no-optimized-zero-stride-load,no-default-unroll,...
    transform_info: Optional[Dict[str, Optional[Union[bool, int]]]] = None
    # options: see ttiimpl_notes.txt
    # TODO: processor/pipeline/mcpu/tune -> ProcessorSettings
    legalization: Optional[Dict[str, RISCVLegalizerSettings]] = None


@dataclass
class PatternGenSettings(YAMLSettings):
    integrated: Optional[bool] = None
    clone_url: Optional[str] = None
    ref: Optional[str] = None
    clone_depth: Optional[int] = None
    sparse_checkout: Optional[bool] = None


@dataclass
class ToolsSettings(YAMLSettings):
    pattern_gen: Optional[PatternGenSettings] = None


@dataclass
class Seal5Settings(YAMLSettings):
    directory: Optional[str] = None
    name: Optional[str] = None
    meta_dir: Optional[str] = None
    logging: Optional[LoggingSettings] = None
    filter: Optional[FilterSettings] = None
    llvm: Optional[LLVMSettings] = None
    git: Optional[GitSettings] = None
    patches: Optional[List[PatchSettings]] = None
    # transform: Optional[TransformSettings] = None  # TODO: make list?
    passes: Optional[PassesSettings] = None  # TODO: make list?
    test: Optional[TestSettings] = None
    extensions: Optional[Dict[str, ExtensionsSettings]] = None
    groups: Optional[GroupsSettings] = None  # TODO: make list?
    inputs: Optional[List[str]] = None
    riscv: Optional[RISCVSettings] = None
    tools: Optional[ToolsSettings] = None
    metrics: list = field(default_factory=list)

    def reset(self):
        self.extensions = {}
        self.patches = []
        self.name = "default"
        self.inputs = []
        self.metrics = []
        self.test = TestSettings(paths=[])
        self.filter = FilterSettings(
            sets=FilterSetting(keep=[], drop=[]),
            instructions=FilterSetting(keep=[], drop=[]),
            aliases=FilterSetting(keep=[], drop=[]),
            intrinsics=FilterSetting(keep=[], drop=[]),
            opcodes=FilterSetting(keep=[], drop=[]),
            encoding_sizes=FilterSetting(keep=[], drop=[]),
        )
        self.passes = PassesSettings(
            defaults=PassesSetting(
                skip=[],
                only=[],
                overrides={},
            ),
            per_model={},
        )
        self.riscv = RISCVSettings(
            xlen=None,
            features=None,
            transform_info=None,
            legalization=None,
        )

    def save(self, dest: Optional[Path] = None):
        if dest is None:
            dest = self.settings_file
        self.to_yaml_file(dest)

    def add_patch(self, patch_settings: PatchSettings):
        for ps in self.patches:
            if ps.name == patch_settings.name:
                raise RuntimeError(f"Duplicate patch '{ps.name}'. Either clean patches or rename patch.")
        self.patches.append(patch_settings)

    @property
    def model_names(self):
        return [Path(path).stem for path in self.inputs]

    @property
    def _meta_dir(self):
        return Path(self.meta_dir)

    @property
    def settings_file(self):
        return self._meta_dir / "settings.yml"

    @property
    def deps_dir(self):
        return self._meta_dir / "deps"

    @property
    def build_dir(self):
        return self._meta_dir / "build"

    @property
    def install_dir(self):
        return self._meta_dir / "install"

    @property
    def logs_dir(self):
        return self._meta_dir / "logs"

    @property
    def models_dir(self):
        return self._meta_dir / "models"

    @property
    def inputs_dir(self):
        return self._meta_dir / "inputs"

    @property
    def temp_dir(self):
        return self._meta_dir / "temp"

    @property
    def gen_dir(self):
        return self._meta_dir / "gen"

    @property
    def tests_dir(self):
        return self._meta_dir / "tests"

    @property
    def patches_dir(self):  # TODO: maybe merge with gen_dir
        return self._meta_dir / "patches"

    @property
    def log_file_path(self):
        return self.logs_dir / "seal5.log"



#
# Copyright (c) 2023 TUM Department of Electrical and Computer Engineering.
#
# This file is part of Seal5.
# See https://github.com/tum-ei-eda/seal5.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Command line subcommand for transforming Seal5 inputs .
    Transform inputs
      1. Create M2-ISA-R metamodel
      2. Convert to Seal5 metamodel (including aliases, builtins,...)
      3. Analyse/optimize instructions
"""

from seal5.flow import Seal5Flow
from seal5.logging import get_logger
from os import getenv


logger = get_logger()


def add_transform_options(parser):
    transform_parser = parser.add_argument_group("transform options")
    transform_parser.add_argument(
        "--skip",
        nargs="+",
        type=str,
        default=None,
        help="Passes that should be skipped",
    )
    transform_parser.add_argument(
        "--only",
        nargs="+",
        type=str,
        default=None,
        help="Passes that should be carried out",
    )


def get_parser(subparsers):
    """ "Define and return a subparser for the transform subcommand."""
    parser = subparsers.add_parser("transform", description="transform Seal5 models.")
    parser.set_defaults(func=handle)
    add_transform_options(parser)
    return parser


def handle(args):
    """Callback function which will be called to process the transform subcommand"""
    if args.dir is None:
        home_dir = getenv("SEAL5_HOME")
        if home_dir is not None:
            args.dir = home_dir
        else:
            logger.error("Seal5_HOME Env var not specified !!!")
    seal5_flow = Seal5Flow(args.dir, name=args.name)
    seal5_flow.transform(
        verbose=args.verbose,
        skip=None if args.skip is None else list(args.skip),
        only=None if args.only is None else list(args.only),
    )



# SPDX-License-Identifier: Apache-2.0
#
# This file is part of the M2-ISA-R project: https://github.com/tum-ei-eda/M2-ISA-R
#
# Copyright (C) 2022
# Chair of Electrical Design Automation
# Technical University of Munich

"""Remove (rd != 0) checks from M2-ISA-R/Seal5 metamodel."""

import sys
import argparse
import logging
import pathlib
import pickle
from typing import Union

from m2isar.metamodel import arch, patch_model

from . import visitor

logger = logging.getLogger("eliminate_mod_rfs")


def get_parser():
    # read command line args
    parser = argparse.ArgumentParser()
    parser.add_argument("top_level", help="A .m2isarmodel or .seal5model file.")
    parser.add_argument("--log", default="info", choices=["critical", "error", "warning", "info", "debug"])
    parser.add_argument("--output", "-o", type=str, default=None)
    return parser


def run(args):
    # initialize logging
    logging.basicConfig(level=getattr(logging, args.log.upper()))

    # resolve model paths
    top_level = pathlib.Path(args.top_level)

    is_seal5_model = False
    if args.output is None:  # inplace
        assert top_level.suffix in [".m2isarmodel", ".seal5model"], "Can not infer model type from file extension."
        if top_level.suffix == ".seal5model":
            is_seal5_model = True

        model_path = top_level
    else:
        model_path = pathlib.Path(args.output)

    logger.info("loading models")
    if not is_seal5_model:
        raise NotImplementedError

    # load models
    with open(top_level, "rb") as f:
        # models: "dict[str, arch.CoreDef]" = pickle.load(f)
        if is_seal5_model:
            model: "dict[str, Union[arch.InstructionSet, ...]]" = pickle.load(f)
            model["cores"] = {}
        else:  # TODO: core vs. set!
            temp: "dict[str, Union[arch.InstructionSet, arch.CoreDef]]" = pickle.load(f)
            assert len(temp) > 0, "Empty model!"
            if isinstance(list(temp.values())[0], arch.CoreDef):
                model = {"cores": temp, "sets": {}}
            elif isinstance(list(temp.values())[0], arch.InstructionSet):
                model = {"sets": temp, "cores": {}}
            else:
                assert False

    for set_name, set_def in model["sets"].items():
        logger.debug("eliminating op mod RFS for set %s", set_def.name)
        patch_model(visitor)
        # TODO: handle RFS symbolically
        for instr_name, instr_def in set_def.instructions.items():
            logger.debug("eliminating op mod RFS for instr %s", instr_def.name)
            instr_def.operation.generate(None)

    logger.info("dumping model")
    with open(model_path, "wb") as f:
        if is_seal5_model:
            pickle.dump(model, f)
        else:
            if len(model["sets"]) > 0:
                assert len(model["cores"]) == 0
                pickle.dump(model["sets"], f)
            elif len(model["cores"]) > 0:
                assert len(model["sets"]) == 0
                pickle.dump(model["cores"], f)
            else:
                assert False


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    run(args)


if __name__ == "__main__":
    main(sys.argv[1:])



# SPDX-License-Identifier: Apache-2.0
#
# This file is part of the M2-ISA-R project: https://github.com/tum-ei-eda/M2-ISA-R
#
# Copyright (C) 2022
# Chair of Electrical Design Automation
# Technical University of Munich

"""Clean M2-ISA-R/Seal5 metamodel to .core_desc file."""

import argparse
import logging
import pathlib
import pickle
from typing import Union

import yaml

from m2isar.metamodel import arch

logger = logging.getLogger("yaml_writer")

# @dataclass
# class Seal5Settings(YAMLSettings):
#     extensions: Optional[Dict[str, ExtensionsSettings]] = None
# @dataclass
# class ExtensionsSettings(YAMLSettings):
#     feature: Optional[str] = None
#     arch: Optional[str] = None
#     version: Optional[str] = None
#     experimental: Optional[bool] = None
#     vendor: Optional[bool] = None
#     instructions: Optional[List[str]] = None
#     NEW: model
#     # patches


def main():
    """Main app entrypoint."""

    # read command line args
    parser = argparse.ArgumentParser()
    parser.add_argument("top_level", help="A .m2isarmodel or .seal5model file.")
    parser.add_argument("--log", default="info", choices=["critical", "error", "warning", "info", "debug"])
    parser.add_argument("--output", "-o", type=str, default=None)
    args = parser.parse_args()

    # initialize logging
    logging.basicConfig(level=getattr(logging, args.log.upper()))

    # resolve model paths
    top_level = pathlib.Path(args.top_level)
    # abs_top_level = top_level.resolve()

    is_seal5_model = False
    # print("top_level", top_level)
    # print("suffix", top_level.suffix)
    if top_level.suffix == ".seal5model":
        is_seal5_model = True
    if args.output is None:
        assert top_level.suffix in [".m2isarmodel", ".seal5model"], "Can not infer model type from file extension."

        out_path = top_level.parent / (top_level.stem + ".core_desc")
    else:
        out_path = pathlib.Path(args.output)

    logger.info("loading models")
    if not is_seal5_model:
        raise NotImplementedError

    # load models
    with open(top_level, "rb") as f:
        # models: "dict[str, arch.CoreDef]" = pickle.load(f)
        if is_seal5_model:
            model: "dict[str, Union[arch.InstructionSet, ...]]" = pickle.load(f)
            model["cores"] = {}
        else:  # TODO: core vs. set!
            temp: "dict[str, Union[arch.InstructionSet, arch.CoreDef]]" = pickle.load(f)
            assert len(temp) > 0, "Empty model!"
            if isinstance(list(temp.values())[0], arch.CoreDef):
                model = {"cores": temp, "sets": {}}
            elif isinstance(list(temp.values())[0], arch.InstructionSet):
                model = {"sets": temp, "cores": {}}
            else:
                assert False

    # preprocess model
    # print("model", model)
    data = {"extensions": {}}
    for set_name, set_def in model["sets"].items():
        # print("set", set_def)
        set_data = {"instructions": []}
        set_data["model"] = top_level.stem
        for instr in set_def.instructions.values():
            set_data["instructions"].append(instr.name)
        data["extensions"][set_name] = set_data
    with open(out_path, "w") as f:
        yaml.dump(data, f)


if __name__ == "__main__":
    main()



# SPDX-License-Identifier: Apache-2.0
#
# This file is part of the M2-ISA-R project: https://github.com/tum-ei-eda/M2-ISA-R
#
# Copyright (C) 2022
# Chair of Electrical Design Automation
# Technical University of Munich

"""Clean M2-ISA-R/Seal5 metamodel to .core_desc file."""

import os
import argparse
import logging
import pathlib
import pickle
from typing import Union

from m2isar.metamodel import arch

from seal5.tools import cdsl2llvm
from seal5.model import Seal5InstrAttribute

logger = logging.getLogger("llvmir_behavior_writer")


def main():
    """Main app entrypoint."""

    # read command line args
    parser = argparse.ArgumentParser()
    parser.add_argument("top_level", help="A .m2isarmodel or .seal5model file.")
    parser.add_argument("--log", default="info", choices=["critical", "error", "warning", "info", "debug"])
    parser.add_argument("--output", "-o", type=str, default=None)
    parser.add_argument("--splitted", action="store_true", help="Split per set and instruction")
    parser.add_argument("--metrics", default=None, help="Output metrics to file")
    parser.add_argument("--ext", type=str, default="ll", help="Default file extension (if using --splitted)")
    args = parser.parse_args()

    # initialize logging
    logging.basicConfig(level=getattr(logging, args.log.upper()))

    # resolve model paths
    top_level = pathlib.Path(args.top_level)
    # abs_top_level = top_level.resolve()

    is_seal5_model = False
    # print("top_level", top_level)
    # print("suffix", top_level.suffix)
    if top_level.suffix == ".seal5model":
        is_seal5_model = True
    if args.output is None:
        assert top_level.suffix in [".m2isarmodel", ".seal5model"], "Can not infer model type from file extension."
        raise NotImplementedError

        # out_path = top_level.parent / (top_level.stem + ".core_desc")
    else:
        out_path = pathlib.Path(args.output)

    logger.info("loading models")
    if not is_seal5_model:
        raise NotImplementedError

    # load models
    with open(top_level, "rb") as f:
        # models: "dict[str, arch.CoreDef]" = pickle.load(f)
        if is_seal5_model:
            model: "dict[str, Union[arch.InstructionSet, ...]]" = pickle.load(f)
            model["cores"] = {}
        else:  # TODO: core vs. set!
            temp: "dict[str, Union[arch.InstructionSet, arch.CoreDef]]" = pickle.load(f)
            assert len(temp) > 0, "Empty model!"
            if isinstance(list(temp.values())[0], arch.CoreDef):
                model = {"cores": temp, "sets": {}}
            elif isinstance(list(temp.values())[0], arch.InstructionSet):
                model = {"sets": temp, "cores": {}}
            else:
                assert False

    # preprocess model
    # print("model", model)
    metrics = {
        "n_sets": 0,
        "n_instructions": 0,
        "n_skipped": 0,
        "n_failed": 0,
        "n_success": 0,
    }
    settings = model.get("settings", None)
    if args.splitted:
        # errs = []
        # model_includes = []
        default_mattr = "+m,+fast-unaligned-access"
        if settings:
            riscv_settings = settings.riscv
            if riscv_settings:
                features = riscv_settings.features
                if features is None:
                    pass
                else:
                    default_mattr = ",".join([f"+{f}" for f in features])
        # errs = []
        assert out_path.is_dir(), "Expecting output directory when using --splitted"
        for set_name, set_def in model["sets"].items():
            xlen = set_def.xlen
            metrics["n_sets"] += 1
            ext_settings = set_def.settings
            for instr_def in set_def.instructions.values():
                metrics["n_instructions"] += 1
                attrs = instr_def.attributes
                if len(attrs) > 0:
                    skip = False
                    if Seal5InstrAttribute.MAY_LOAD in attrs:
                        skip = True
                    elif Seal5InstrAttribute.MAY_STORE in attrs:
                        skip = True
                    elif arch.InstrAttribute.COND in attrs:
                        skip = True
                    elif arch.InstrAttribute.NO_CONT in attrs:
                        skip = True
                    if skip:
                        metrics["n_skipped"] += 1
                        continue
                input_file = out_path / set_name / f"{instr_def.name}.core_desc"
                if not input_file.is_file():
                    metrics["n_skipped"] += 1
                    # errs.append(TODO)
                output_file = out_path / set_name / f"{instr_def.name}.{args.ext}"
                install_dir = os.getenv("CDSL2LLVM_DIR", None)
                assert install_dir is not None
                install_dir = pathlib.Path(install_dir)
                mattr = default_mattr
                if ext_settings is not None:
                    # predicate = ext_settings.get_predicate(name=set_name)
                    arch_ = ext_settings.get_arch(name=set_name)
                    mattr = ",".join([*mattr.split(","), f"+{arch_}"])
                if xlen == 64 and "+64bit" not in mattr:
                    mattr = ",".join([*mattr.split(","), "+64bit"])
                try:
                    cdsl2llvm.run_pattern_gen(
                        # install_dir / "llvm" / "build",
                        install_dir,
                        input_file,
                        output_file,
                        skip_patterns=True,
                        skip_formats=True,
                        mattr=mattr,
                        xlen=xlen,
                    )
                    metrics["n_success"] += 1
                except AssertionError:
                    pass
                    metrics["n_failed"] += 1
                    # errs.append((insn_name, str(ex)))
        # if len(errs) > 0:
        #     # print("errs", errs)
        #     for insn_name, err_str in errs:
        #         print("Err:", insn_name, err_str)
        #         input("!")
        #         cdsl2llvm.
        #         out_path_ = out_path / set_name / f"{instr_def.name}.{args.ext}"
        #         out_path_.parent.mkdir(exist_ok=True)
    else:
        raise NotImplementedError
    if args.metrics:
        metrics_file = args.metrics
        with open(metrics_file, "w") as f:
            f.write(",".join(metrics.keys()))
            f.write("\n")
            f.write(",".join(map(str, metrics.values())))
            f.write("\n")


if __name__ == "__main__":
    main()



# SPDX-License-Identifier: Apache-2.0
#
# This file is part of the M2-ISA-R project: https://github.com/tum-ei-eda/M2-ISA-R
#
# Copyright (C) 2022
# Chair of Electrical Design Automation
# Technical University of Munich

"""Detect available registers for Seal5."""

import sys
import argparse
import logging
import pathlib
import pickle
from typing import Union

from m2isar.metamodel import arch
import seal5.model as seal5_model

logger = logging.getLogger("detect_registers")


def detect_registers(set_def: seal5_model.Seal5InstructionSet):
    to_skip = ["FENCE", "RES", "PRIV", "DPC"]  # TODO: eliminate unused memories during optimization?
    for mem_name, mem in set_def.memories.items():
        if mem_name in to_skip:
            continue
        if mem.is_main_mem or mem.is_pc:
            continue
        assert mem_name not in set_def.registers, "Registers already added to set."
        name = mem_name
        size = mem.size
        width = None
        reg_class = seal5_model.Seal5RegisterClass.CUSTOM
        if name == "X":
            reg_class = seal5_model.Seal5RegisterClass.GPR
        elif name == "F":
            reg_class = seal5_model.Seal5RegisterClass.FPR
        elif name == "CSR":
            reg_class = seal5_model.Seal5RegisterClass.CSR
        if size is None:
            raise NotImplementedError(f"Register with unknown size: {name}")
        if size > 1:  # TODO: check this!
            names = [f"{name}{i}" for i in range(size)]
            group = seal5_model.Seal5RegisterGroup(names, size, width, reg_class)
            set_def.register_groups[name] = group
            for reg in group.registers:
                set_def.registers[name] = reg
        else:
            reg = seal5_model.Seal5Register(name, size, width, reg_class)
            set_def.registers[name] = reg


def get_parser():
    # read command line args
    parser = argparse.ArgumentParser()
    parser.add_argument("top_level", help="A .m2isarmodel or .seal5model file.")
    parser.add_argument("--log", default="info", choices=["critical", "error", "warning", "info", "debug"])
    parser.add_argument("--output", "-o", type=str, default=None)
    return parser


def run(args):
    # initialize logging
    logging.basicConfig(level=getattr(logging, args.log.upper()))

    # resolve model paths
    top_level = pathlib.Path(args.top_level)

    if args.output is None:  # inplace
        assert top_level.suffix == ".seal5model", "Can not infer model type from file extension."
        model_path = top_level
    else:
        model_path = pathlib.Path(args.output)

    logger.info("loading models")

    # load models
    with open(top_level, "rb") as f:
        # models: "dict[str, arch.CoreDef]" = pickle.load(f)
        model: "dict[str, Union[arch.InstructionSet, ...]]" = pickle.load(f)

    to_skip = ["Zicsr"]  # TODO: use yaml!
    for set_name, set_def in model["sets"].items():
        if set_name in to_skip:
            continue
        logger.info("processing set %s", set_name)
        detect_registers(set_def)

    logger.info("dumping model")
    with open(model_path, "wb") as f:
        pickle.dump(model, f)


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    run(args)


if __name__ == "__main__":
    main(sys.argv[1:])
